#!/usr/bin/env python3
#
#  repo.py
"""
Modified Dulwich repository object.

.. versionadded:: 0.3.0
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#
#  get_user_identity and Repo based on https://github.com/dulwich/dulwich
#  Copyright (C) 2013 Jelmer Vernooij <jelmer@jelmer.uk>
#  |  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  |  not use this file except in compliance with the License. You may obtain
#  |  a copy of the License at
#  |
#  |	  http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  |  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  |  License for the specific language governing permissions and limitations
#  |  under the License.
#

# stdlib
import os
from itertools import chain
from typing import Any, Dict, Iterator, Optional, Type, TypeVar, Union, cast

# 3rd party
import click
import dulwich.index
from domdf_python_tools.paths import PathPlus
from domdf_python_tools.typing import PathLike
from dulwich import repo
from dulwich.config import StackedConfig
from dulwich.objects import Commit, Tree, TreeEntry

__all__ = ["get_user_identity", "Repo", "_R"]

_R = TypeVar("_R", bound="Repo")


def get_user_identity(config: StackedConfig, kind: Optional[str] = None) -> bytes:
	"""
	Determine the identity to use for new commits.

	If kind is set, this first checks
	:envvar:`GIT_${KIND}_NAME` and :envvar:`GIT_${KIND}_EMAIL`.

	If those variables are not set, then it will fall back
	to reading the ``user.name`` and ``user.email`` settings from
	the specified configuration.

	If that also fails, then it will fall back to using
	the current users' identity as obtained from the host
	system (e.g. the gecos field, $EMAIL, $USER@$(hostname -f).

	:param config:
	:param kind: Optional kind to return identity for, usually either ``'AUTHOR'`` or ``'COMMITTER'``.

	:returns: A user identity
	"""

	user: Optional[bytes] = None
	email: Optional[bytes] = None

	if kind:
		user_uc = os.environ.get("GIT_" + kind + "_NAME")
		if user_uc is not None:
			user = user_uc.encode("UTF-8")
		email_uc = os.environ.get("GIT_" + kind + "_EMAIL")
		if email_uc is not None:
			email = email_uc.encode("UTF-8")

	if user is None:
		try:
			user = config.get(("user", ), "name")
		except KeyError:
			user = None

	if email is None:
		try:
			email = config.get(("user", ), "email")
		except KeyError:
			email = None

	if user is None or email is None:
		default_user, default_email = repo._get_default_identity()  # type: ignore

		if user is None:
			user = default_user.encode("UTF-8")
		if email is None:
			email = default_email.encode("UTF-8")

	if email.startswith(b'<') and email.endswith(b'>'):
		email = email[1:-1]

	return user + b" <" + email + b">"


class Repo(repo.Repo):
	"""
	Modified Dulwich repository object.

	A git repository backed by local disk.

	To open an existing repository, call the constructor with
	the path of the repository.

	To create a new repository, use the Repo.init class method.

	:param root:
	"""

	def do_commit(
			self,
			message: Optional[Union[str, bytes]] = None,
			committer: Optional[Union[str, bytes]] = None,
			author: Optional[Union[str, bytes]] = None,
			commit_timestamp: Optional[float] = None,
			commit_timezone: Optional[float] = None,
			author_timestamp: Optional[float] = None,
			author_timezone: Optional[float] = None,
			tree: Optional[Any] = None,
			encoding: Optional[Union[str, bytes]] = None,
			ref: bytes = b'HEAD',
			merge_heads: Optional[Any] = None
			) -> bytes:
		"""
		Create a new commit.

		If not specified, `committer` and `author` default to
		:func:`get_user_identity(..., 'COMMITTER') <.get_user_identity>`
		and :func:`get_user_identity(..., 'AUTHOR') <.get_user_identity>` respectively.

		:param message: Commit message
		:param committer: Committer fullname
		:param author: Author fullname
		:param commit_timestamp: Commit timestamp (defaults to now)
		:param commit_timezone: Commit timestamp timezone (defaults to GMT)
		:param author_timestamp: Author timestamp (defaults to commit timestamp)
		:param author_timezone: Author timestamp timezone (defaults to commit timestamp timezone)
		:param tree: SHA1 of the tree root to use (if not specified the current index will be committed).
		:param encoding: Encoding
		:param ref: Optional ref to commit to (defaults to current branch)
		:param merge_heads: Merge heads (defaults to .git/MERGE_HEADS)

		:returns: New commit SHA1
		"""

		config = self.get_config_stack()

		if committer is None:
			committer = get_user_identity(config, kind="COMMITTER")

		if author is None:
			try:
				author = get_user_identity(config, kind="AUTHOR")
			except ModuleNotFoundError as e:
				if str(e) == "No module named 'pwd'":
					author = committer
				else:
					raise

		return super().do_commit(
				message=message,
				committer=committer,
				author=author,
				commit_timestamp=commit_timestamp,
				commit_timezone=commit_timezone,
				author_timestamp=author_timestamp,
				author_timezone=author_timezone,
				tree=tree,
				encoding=encoding,
				ref=ref,
				merge_heads=merge_heads,
				)

	def _get_user_identity(
			self,
			config: "StackedConfig",
			kind: Optional[str] = None,
			) -> bytes:
		"""
		Determine the identity to use for new commits.

		:param config:
		:param kind:
		"""

		return get_user_identity(config)

	@classmethod
	def init(cls: Type[_R], path: PathLike, mkdir: bool = False) -> _R:
		"""
		Create a new repository.

		:param path: Path in which to create the repository.
		:param mkdir: Whether to create the directory if it doesn't exist.
		"""

		return super().init(path, mkdir)

	@classmethod
	def init_bare(cls: Type[_R], path: PathLike, mkdir: bool = False) -> _R:
		"""
		Create a new bare repository.

		:param path: Path in which to create the repository.
		:param mkdir:
		"""

		return super().init_bare(path, mkdir)

	def list_remotes(self) -> Dict[str, str]:
		"""
		Returns a mapping of remote names to remote URLs, for the repo's current remotes.

		.. versionadded:: 0.7.0
		"""

		remotes = {}
		config = self.get_config()

		for key in list(config.keys()):
			if key[0] == b"remote":
				remotes[key[1].decode("UTF-8")] = config.get(key, "url").decode("UTF-8")

		return remotes

	def reset_to(self, sha: Union[str, bytes]):
		"""
		Reset the state of the repository to the given commit sha.

		Any files added in subsequent commits will be removed,
		any deleted will be restored,
		and any modified will be reverted.

		.. versionadded:: 0.8.0

		:param sha:
		"""

		# this package
		from southwark import status

		if isinstance(sha, str):
			sha = sha.encode("UTF-8")

		index = self.open_index()
		directory = PathPlus(self.path)

		tree_for_sha: Tree = cast(Commit, self[sha]).tree
		tree_for_head: Tree = cast(Commit, self[b'HEAD']).tree

		dulwich.index.build_index_from_tree(
				root_path=directory,
				index_path=self.index_path(),
				object_store=self.object_store,
				tree_id=tree_for_sha,
				)

		try:
			# Based on https://github.com/dulwich/dulwich/issues/588#issuecomment-348412641
			oldtree: Tree = cast(Tree, self[tree_for_head])
			newtree: Tree = cast(Tree, self[tree_for_sha])

			contents_iterator: Iterator[TreeEntry] = self.object_store.iter_tree_contents(newtree.id)
			desired_filenames = [f.path for f in contents_iterator]

			for f in self.object_store.iter_tree_contents(oldtree.id):
				if f.path not in desired_filenames:
					# delete files that were in old branch, but not new
					(directory / f.path.decode("UTF-8")).unlink()

		except KeyError:
			click.echo("Unable to delete filed added in later commits", err=True)

		self[b'HEAD'] = sha
		index.write()

		current_status = status(self)

		for filename in chain.from_iterable([
				current_status.staged["add"],
				current_status.staged["delete"],
				current_status.staged["modify"],
				]):
			self.stage(os.path.normpath(filename.as_posix()))



#!/usr/bin/env python3
#
#  __init__.py
"""
Extensions to the Dulwich Git library.
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#
#  format_commit, get_untracked_paths, get_tree_changes, clone and status
#  based on https://github.com/dulwich/dulwich
#  Copyright (C) 2013 Jelmer Vernooij <jelmer@jelmer.uk>
#  |  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  |  not use this file except in compliance with the License. You may obtain
#  |  a copy of the License at
#  |
#  |      http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  |  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  |  License for the specific language governing permissions and limitations
#  |  under the License.
#

# stdlib
import os
import shutil
from contextlib import closing, contextmanager
from itertools import chain
from operator import itemgetter
from typing import (
		IO,
		ContextManager,
		Dict,
		Iterator,
		List,
		NamedTuple,
		Optional,
		Sequence,
		Tuple,
		TypeVar,
		Union,
		overload
		)

# 3rd party
import dulwich.repo
from click import echo
from consolekit.terminal_colours import Fore
from domdf_python_tools.compat import nullcontext
from domdf_python_tools.paths import PathPlus, maybe_make
from domdf_python_tools.typing import PathLike
from dulwich.config import StackedConfig
from dulwich.ignore import IgnoreFilterManager
from dulwich.index import Index, get_unstaged_changes
from dulwich.objects import Commit, ShaFile, Tag
from dulwich.porcelain import default_bytes_err_stream, fetch, path_to_tree_path
from typing_extensions import TypedDict

# this package
from southwark.repo import Repo

__author__: str = "Dominic Davis-Foster"
__copyright__: str = "2020 Dominic Davis-Foster"
__license__: str = "MIT License"
__version__: str = "0.8.0"
__email__: str = "dominic@davis-foster.co.uk"

__all__ = [
		"get_tags",
		"check_git_status",
		"assert_clean",
		"get_untracked_paths",
		"status",
		"StagedDict",
		"GitStatus",
		"get_tree_changes",
		"clone",
		"_DR",
		"open_repo_closing",
		"windows_clone_helper",
		]

_DR = TypeVar("_DR", bound=dulwich.repo.Repo)


class StagedDict(TypedDict):
	"""
	The values are lists of filenames, relative to the repository root.

	.. versionadded:: 0.6.1
	"""

	add: List[PathPlus]
	delete: List[PathPlus]
	modify: List[PathPlus]


class GitStatus(NamedTuple):
	"""
	Represents the output of :func:`~.status`.

	.. versionadded:: 0.6.1
	"""

	#: Dict with lists of staged paths.
	staged: StagedDict

	#: List of unstaged paths.
	unstaged: List[PathPlus]

	#: List of untracked, un-ignored & non-.git paths.
	untracked: List[PathPlus]


def get_tags(repo: Union[dulwich.repo.Repo, PathLike] = '.') -> Dict[str, str]:
	"""
	Returns a mapping of commit SHAs to tags.

	:param repo:
	"""

	tags: Dict[str, str] = {}

	with open_repo_closing(repo) as r:
		raw_tags: Dict[bytes, bytes] = r.refs.as_dict(b"refs/tags")
		for tag, sha, in raw_tags.items():
			obj = r.get_object(sha)
			if isinstance(obj, Tag):
				tags[obj.object[1].decode("UTF-8")] = tag.decode("UTF-8")
			elif isinstance(obj, Commit):
				tags[sha.decode("UTF-8")] = tag.decode("UTF-8")

	return tags


def assert_clean(repo: PathPlus, allow_config: Sequence[PathLike] = ()) -> bool:
	"""
	Returns :py:obj:`True` if the working directory is clean.

	If not, returns :py:obj:`False` and prints a helpful error message to stderr.

	:param repo:
	:param allow_config:
	"""

	allow_config = [PathPlus(filename) for filename in allow_config]

	stat = status(repo)

	modified_files = chain.from_iterable([
			stat.staged["add"],
			stat.staged["delete"],
			stat.staged["modify"],
			stat.unstaged,
			])

	if modified_files:
		for filename in modified_files:
			if filename not in allow_config:
				break
		else:
			return True

	# If we get to here the directory isn't clean
	echo(Fore.RED("Git working directory is not clean:"), err=True)

	for line in format_git_status(stat):
		echo(Fore.RED(f"  {line}"), err=True)

	return False


status_codes: Dict[str, str] = {
		"add": 'A',
		"delete": 'D',
		"modify": 'M',
		}


def check_git_status(repo_path: PathLike) -> Tuple[bool, List[str]]:
	"""
	Check the ``git`` status of the given repository.

	:param repo_path: Path to the repository root.

	:return: Whether the git working directory is clean, and the list of uncommitted files if it isn't.
	"""

	str_lines = list(format_git_status(status(repo_path)))

	# with in_directory(repo_path):
	#
	# 	lines = [
	# 			line.strip()
	# 			for line in subprocess.check_output(["git", "status", "--porcelain"]).splitlines()
	# 			if not line.strip().startswith(b"??")
	# 			]
	#
	# str_lines = [line.decode("UTF-8") for line in lines]
	return not bool(str_lines), str_lines


def format_git_status(status: GitStatus) -> Iterator[str]:
	"""
	Format the ``git`` status of the given repository for output to the terminal.

	:param status:

	:return: An iterator over the formatted list of uncommitted files.

	.. versionadded:: 0.6.1
	"""

	files: Dict[bytes, str] = {}

	for key, code in status_codes.items():
		for file in status.staged[key]:  # type: ignore
			if file in files:
				files[file] += code
			else:
				files[file] = code

	for file in status.unstaged:
		if file in files:
			files[file] += 'M'
		else:
			files[file] = 'M'

	for file, codes in sorted(files.items(), key=itemgetter(0)):
		longest = max(len(v) for v in files.values()) + 1

		status_code = ''.join(sorted(codes)).ljust(longest, ' ')
		yield f"{status_code}{file!s}"


status_excludes = {".git", ".tox", ".tox4", ".mypy_cache", ".pytest_cache", "venv", ".venv"}


def get_untracked_paths(path: PathLike, index: Index) -> Iterator[str]:
	"""
	Returns a list of untracked files.

	:param path: Path to walk.
	:param index: Index to check against.
	"""

	path = str(path)

	for dirpath, dirnames, filenames in os.walk(path):
		# Skip .git etc. and below.
		for exclude in status_excludes:
			if exclude in dirnames:
				dirnames.remove(exclude)
				if dirpath != path:
					continue
			if exclude in filenames:
				filenames.remove(exclude)
				if dirpath != path:
					continue

		for filename in filenames:
			filepath = os.path.join(dirpath, filename)

			_pp_filename = (PathPlus(path) / filepath)
			if _pp_filename.is_symlink() and not _pp_filename.resolve().is_relative_to(path):
				continue

			ip = path_to_tree_path(path, filepath)

			if ip not in index:
				yield os.path.relpath(filepath, path)


def get_tree_changes(repo: Union[PathLike, dulwich.repo.Repo]) -> StagedDict:
	"""
	Return add/delete/modify changes to tree by comparing the index to HEAD.

	:param repo: repo path or object.

	:returns: Dictionary containing changes for each type of change.

	.. versionadded:: 0.6.1
	"""

	with open_repo_closing(repo) as r:
		index = r.open_index()

		# Compares the Index to the HEAD & determines changes
		# Iterate through the changes and report add/delete/modify
		# TODO: call out to dulwich.diff_tree somehow.
		tracked_changes: StagedDict = {
				"add": [],
				"delete": [],
				"modify": [],
				}
		try:
			tree_id = r[b'HEAD'].tree  # type: ignore
		except KeyError:
			tree_id = None

		for change in index.changes_from_tree(r.object_store, tree_id):
			if not change[0][0]:
				tracked_changes["add"].append(PathPlus(change[0][1].decode("UTF-8")))
			elif not change[0][1]:
				tracked_changes["delete"].append(PathPlus(change[0][0].decode("UTF-8")))
			elif change[0][0] == change[0][1]:
				tracked_changes["modify"].append(PathPlus(change[0][0].decode("UTF-8")))
			else:
				raise NotImplementedError("git mv ops not yet supported")
		return tracked_changes


def status(repo: Union[dulwich.repo.Repo, PathLike] = '.') -> GitStatus:
	"""
	Returns staged, unstaged, and untracked changes relative to the HEAD.

	:param repo: Path to repository or repository object.
	"""

	with open_repo_closing(repo) as r:
		# 1. Get status of staged
		tracked_changes = get_tree_changes(r)

		# 2. Get status of unstaged
		index = r.open_index()
		normalizer = r.get_blob_normalizer()
		filter_callback = normalizer.checkin_normalize
		unstaged_changes = [
				PathPlus(p.decode("UTF-8")) for p in get_unstaged_changes(index, str(r.path), filter_callback)
				]

		# Remove ignored files
		ignore_manager = IgnoreFilterManager.from_repo(r)
		untracked_changes = [
				PathPlus(p) for p in get_untracked_paths(r.path, index) if not ignore_manager.is_ignored(p)
				]

		return GitStatus(tracked_changes, unstaged_changes, untracked_changes)


def clone(
		source: Union[str, bytes],
		target: Union[PathLike, bytes, None] = None,
		bare: bool = False,
		checkout: Optional[bool] = None,
		errstream: IO = default_bytes_err_stream,
		origin: Union[str, bytes] = "origin",
		depth: Optional[int] = None,
		**kwargs,
		) -> Repo:
	"""
	Clone a local or remote git repository.

	:param source: Path or URL for source repository.
	:param target: Path to target repository.
	:param bare: Whether to create a bare repository.
	:param checkout: Whether to check-out HEAD after cloning.
	:param errstream: Optional stream to write progress to.
	:param origin: Name of remote from the repository used to clone.
	:param depth: Depth to fetch at.

	:returns: The cloned repository.

	.. versionadded:: 0.6.1

	.. versionchanged:: 0.7.2

		* ``target`` now accepts :py:data:`domdf_python_tools.types.PathLike` objects.
		* ``origin`` now accepts :class:`str` objects.
	"""

	if checkout is None:
		checkout = (not bare)
	elif checkout and bare:
		raise TypeError("'checkout' and 'bare' are incompatible.")

	if isinstance(origin, bytes):
		origin = origin.decode("UTF-8")

	if isinstance(source, bytes):
		source = source.decode("UTF-8")

	if target is None:
		target = source.split('/')[-1]

	if isinstance(target, bytes):
		target = target.decode("UTF-8")

	maybe_make(target)

	if bare:
		r = Repo.init_bare(target)
	else:
		r = Repo.init(target)

	try:
		target_config = r.get_config()

		target_config.set(("remote", origin), "url", source.encode("UTF-8"))
		target_config.set(("remote", origin), "fetch", f"+refs/heads/*:refs/remotes/{origin}/*".encode("UTF-8"))
		target_config.write_to_path()
		fetch_result = fetch(
				r,
				origin,
				errstream=errstream,
				message=f"clone: from {source}".encode("UTF-8"),
				depth=depth,
				**kwargs,
				)

		head: Optional[ShaFile]

		try:
			head = r[fetch_result.refs[b'HEAD']]
		except KeyError:
			head = None
		else:
			r[b'HEAD'] = head.id

		if checkout and not bare and head is not None:
			errstream.write(b'Checking out ' + head.id + b'\n')
			r.reset_index(head.tree)  # type: ignore   # TODO

	except BaseException:
		shutil.rmtree(target, ignore_errors=True)
		r.close()
		raise

	return r


@overload
def open_repo_closing(path_or_repo: _DR) -> ContextManager[_DR]:
	...  # pragma: no cover


@overload
def open_repo_closing(path_or_repo: Union[str, os.PathLike]) -> ContextManager[Repo]:
	...  # pragma: no cover


def open_repo_closing(path_or_repo):
	"""
	Returns a context manager which will return :class:`dulwich.repo.Repo` objects unchanged,
	but will create a new :class:`dulwich.repo.Repo` when a filesystem path is given.

	.. versionadded:: 0.7.0

	:param path_or_repo: Either a :class:`dulwich.repo.Repo` object or the path of a repository.
	"""  # noqa: D400

	if isinstance(path_or_repo, dulwich.repo.BaseRepo):
		return nullcontext(path_or_repo)

	return closing(Repo(path_or_repo))


@contextmanager
def windows_clone_helper():
	"""
	Contextmanager to aid cloning on Windows during tests.

	.. versionadded:: 0.8.0

	.. attention:: This function is intended only for use in tests.

	Usage:

	.. code-block:: python

		with windows_clone_helper():
			repo = clone(...)

	"""

	_environ = dict(os.environ)  # or os.environ.copy()
	_default_backends = StackedConfig.default_backends

	try:
		name = "wordle_user"
		StackedConfig.default_backends = lambda *args: []  # type: ignore
		os.environ["USER"] = os.environ.get("USER", name)

		yield

	finally:
		os.environ.clear()
		os.environ.update(_environ)
		StackedConfig.default_backends = _default_backends  # type: ignore



#!/usr/bin/env python3
#
#  log.py
"""
Python implementation of ``git log``.
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import time
from datetime import datetime
from textwrap import indent
from typing import Dict, Mapping, Optional, Union

# 3rd party
from consolekit.terminal_colours import Fore, strip_ansi
from domdf_python_tools.stringlist import DelimitedList, StringList
from domdf_python_tools.typing import PathLike
from dulwich.objects import Commit, format_timezone
from dulwich.repo import Repo

# this package
from southwark import get_tags

__all__ = ["Log"]

yellow_meta_left = Fore.YELLOW(" (")
yellow_meta_comma = Fore.YELLOW(", ")
yellow_meta_right = Fore.YELLOW(')')


class Log:
	"""
	Python implementation of ``git log``.

	:param repo: The git repository.
	"""

	#: The git repository.
	repo: Repo

	def __init__(self, repo: Union[Repo, PathLike] = '.'):
		if isinstance(repo, Repo):
			self.repo = repo
		else:
			self.repo = Repo(repo)

		#: Mapping of commit SHAs to tags.
		self.tags = get_tags(self.repo)

		#: Mapping of git refs to commit SHAs.
		self.refs: Dict[str, str] = {
				k.decode("UTF-8"): v.decode("UTF-8")
				for k,
				v in self.repo.get_refs().items()
				if not k.startswith(b"refs/tags/")
				}

		#: Mapping of local branches to the SHA of the latest commit in that branch.
		self.local_branches: Dict[str, str] = {}

		#: Mapping of remote branches to the SHA of the latest commit in that branch.
		self.remote_branches: Dict[str, str] = {}

		#: The name of the current branch
		self.current_branch: str = self.repo.refs.follow(b"HEAD")[0][1].decode("UTF-8")[11:]

		for key, value in self.refs.items():
			if key.startswith("refs/heads/"):
				self.local_branches[key[11:]] = value
			elif key.startswith("refs/remotes/"):
				self.remote_branches[key[13:]] = value

	# Based on https://www.dulwich.io/code/dulwich/blob/master/dulwich/porcelain.py
	def format_commit(self, commit: Commit) -> StringList:
		"""
		Return a human-readable commit log entry.

		:param commit: A `Commit` object
		"""

		buf = StringList()
		meta = []
		commit_id = commit.id.decode("UTF-8")

		if "HEAD" in self.refs and self.refs["HEAD"] == commit_id:
			for branch, sha in self.local_branches.items():
				if sha == commit_id and branch == self.current_branch:
					meta.append(Fore.BLUE("HEAD -> ") + Fore.GREEN(branch))
					break

		if commit_id in self.tags:
			meta.append(Fore.YELLOW(f"tag: {self.tags[commit_id]}"))

		for branch, sha in self.remote_branches.items():
			if sha == commit_id:
				meta.append(Fore.RED(branch))
				break

		if "HEAD" in self.refs and self.refs["HEAD"] == commit_id:
			for branch, sha in self.local_branches.items():
				if sha == commit_id and branch != self.current_branch:
					meta.append(Fore.GREEN(branch))
					break

		if meta:
			meta_string = yellow_meta_left + yellow_meta_comma.join(meta) + yellow_meta_right
		else:
			meta_string = ''

		buf.append(Fore.YELLOW("commit: " + commit.id.decode("UTF-8") + meta_string))

		if len(commit.parents) > 1:
			parents = DelimitedList(c.decode("UTF-8") for c in commit.parents[1:])
			buf.append(f"merge: {parents:...}")

		buf.append("Author: " + commit.author.decode("UTF-8"))

		if commit.author != commit.committer:
			buf.append("Committer: " + commit.committer.decode("UTF-8"))

		time_tuple = time.gmtime(commit.author_time + commit.author_timezone)
		time_str = time.strftime("%a %b %d %Y %H:%M:%S", time_tuple)
		timezone_str = format_timezone(commit.author_timezone).decode("UTF-8")
		buf.append(f"Date:   {time_str} {timezone_str}")

		buf.blankline()
		buf.append(indent(commit.message.decode("UTF-8"), "    "))
		buf.blankline(ensure_single=True)

		return buf

	def log(
			self,
			max_entries: Optional[int] = None,
			reverse: bool = False,
			from_date: Optional[datetime] = None,
			from_tag: Optional[str] = None,
			colour: bool = True
			) -> str:
		"""
		Return the formatted commit log.

		:param max_entries: Maximum number of entries to display
		:default max_entries: all entries
		:param reverse: Print entries in reverse order.
		:param from_date: Show commits after the given date.
		:param from_tag: Show commits after the given tag.
		:param colour: Show coloured output.
		"""

		kwargs: Mapping[str, Union[None, int, bool]] = dict(max_entries=max_entries, reverse=reverse)

		if from_date is not None and from_tag is not None:
			raise ValueError("'from_date' and 'from_tag' are exclusive.")
		elif from_date:
			kwargs["since"] = from_date.timestamp()  # type: ignore
		elif from_tag and not any(from_tag == tag for tag in self.tags.values()):
			raise ValueError(f"No such tag {from_tag!r}")

		buf = StringList()
		walker = self.repo.get_walker(**kwargs)

		for entry in walker:
			buf.append(str(self.format_commit(entry.commit)))

			if from_tag:
				commit_id = entry.commit.id.decode("UTF-8")
				if commit_id in self.tags and self.tags[commit_id] == from_tag:
					if reverse:
						buf = StringList([str(self.format_commit(entry.commit))])
					else:
						break

		if colour:
			return str(buf)
		else:
			return strip_ansi(str(buf))



#!/usr/bin/env python3
#
#  __init__.py
"""
Extensions to `click <https://click.palletsprojects.com>`_.

.. versionadded:: 0.5.0
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
from typing import Callable, Optional

# 3rd party
import click

__all__ = ["commit_option", "commit_message_option"]


def commit_option(default: Optional[bool]) -> Callable:
	"""
	Decorator to add the ``--commit / --no-commit`` option to a click command.

	.. versionadded:: 0.5.0

	:param default: Whether to commit automatically.

	* :py:obj:`None` -- Ask first
	* :py:obj:`True` -- Commit automatically
	* :py:obj:`False` -- Don't commit
	"""

	if default is True:
		default_text = "Commit automatically"
	elif default is False:
		default_text = "Don't commit"
	else:
		default_text = "Ask first"

	return click.option(
			"-y/-n",
			"--commit/--no-commit",
			default=lambda: default,
			show_default=default_text,
			help="Commit or do not commit any changed files.",
			)


def commit_message_option(default: str) -> Callable:
	"""
	Decorator to add the ``-m / --message`` option to a click command.

	.. versionadded:: 0.5.0

	:param default: The default commit message.
	"""

	return click.option(
			"-m",
			"--message",
			type=click.STRING,
			default=default,
			help="The commit message to use.",
			show_default=True,
			)



#!/usr/bin/env python3
#
#  targit.py
"""
Archive where the changes to the contents are recorded using `git <https://git-scm.com/>`_.
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#

# stdlib
import atexit
import datetime
import getpass
import os
import re
import socket
import tarfile
import time
from tempfile import TemporaryDirectory
from typing import Iterator, List, NamedTuple, Optional

# 3rd party
from domdf_python_tools.doctools import prettify_docstrings
from domdf_python_tools.paths import PathPlus
from domdf_python_tools.typing import PathLike
from dulwich.objects import format_timezone
from dulwich.repo import Repo
from filelock import FileLock, Timeout
from typing_extensions import Literal

# this package
from southwark import StagedDict, status

__all__ = [
		"Modes",
		"Status",
		"SaveState",
		"check_archive_paths",
		"BadArchiveError",
		"TarGit",
		]

Modes = Literal["r", "w", "a"]
"""
Valid modes for opening :class:`~.TarGit` archives in

* ``'r'`` -- Read only access. The archive must exist.
* ``'w'`` -- Read and write access. The archive must not exist.
* ``'a'`` -- Read and write access to an existing archive.
"""

Status = StagedDict
"""
Represents the dictionary returned by :meth:`TarGit.status() <.TarGit.status>`.

The values are lists of filenames, relative to the TarGit root.
"""


@prettify_docstrings
class SaveState(NamedTuple):
	"""
	Represents a save event in a :class:`~.TarGit` archive's history.
	"""

	# TODO: changed files

	#: The SHA id of the underlying commit.
	id: str  # noqa: A003  # pylint: disable=redefined-builtin

	#: The name of the user who made the changes.
	user: str

	#: The hostname of the device the changes were made on.
	device: str

	#: The time the changes were saved, in seconds from epoch.
	time: float

	#: The timezone the changes were made in, as a GMT offset in seconds.
	timezone: int

	def format_time(self) -> str:
		"""
		Format the save state's time in the following format::

			Thu Oct 29 2020 15:53:52 +0000

		where ``+0000`` represents GMT.
		"""  # noqa: D400

		time_tuple = time.gmtime(self.time + self.timezone)
		time_str = time.strftime("%a %b %d %Y %H:%M:%S", time_tuple)
		timezone_str = format_timezone(self.timezone).decode("UTF-8")
		return f"{time_str} {timezone_str}"


def check_archive_paths(archive: tarfile.TarFile) -> bool:
	"""
	Checks the contents of an archive to ensure it does not contain
	any filenames with absolute paths or path traversal.

	For example, the following paths would raise an :exc:`~.BadArchiveError`:

	* ``/usr/bin/malware.sh`` -- this is an absolute path.
	* ``~/.local/bin/malware.sh`` -- this tries to put the file in the user's home directory.
	* ``../.local/bin/malware.sh`` -- this uses path traversal to try to get to a parent directory.

	.. seealso:: The warning for :meth:`tarfile.TarFile.extractall` in the Python documentation.

	:param archive:
	"""  # noqa: D400

	for member_name in archive.getnames():
		member_name_p = PathPlus(member_name)
		if member_name_p.is_absolute() or ".." in member_name_p.parts or member_name.startswith('~'):
			raise BadArchiveError

	return True


class BadArchiveError(IOError):
	"""
	Exception to indicate an archive contains files utilising path traversal.
	"""

	def __init__(self):
		super().__init__("Refusing to extract an archive containing files utilising path traversal.")


class TarGit(os.PathLike):
	"""
	A "TarGit" (pronounced "target", /tɑːɡɪt/) is a ``tar.gz`` archive where the changes to the contents are
	recorded using `git <https://git-scm.com/>`_.

	:param filename: The filename of the archive.
	:param mode: The mode to open the file in.

	:raises FileNotFoundError: If the file is opened in read or append mode, but it does not exist.
	:raises FileExistsError: If the file is opened in write mode, but it already exists.
	:raises ValueError: If an unknown value for ``mode`` is given.
	"""  # noqa: D400

	__mode: Modes
	__repo: Repo
	__lock: Optional[FileLock]

	def __init__(self, filename: PathLike, mode: Modes = 'r'):
		self.filename = PathPlus(filename)
		self.__closed: bool = True

		self.__tmpdir: TemporaryDirectory = TemporaryDirectory()
		self.__tmpdir_p = PathPlus(self.__tmpdir.name)
		atexit.register(self.__exit_handler)

		if mode in {'w', 'a'}:
			lock_file = str(self.filename.with_suffix(self.filename.suffix + ".lock"))
			self.__lock = FileLock(lock_file, timeout=1)
			try:
				self.__lock.acquire()
			except Timeout:
				raise OSError(f"Unable to acquire a lock for the file '{self.filename!s}'")
		else:
			self.__lock = None

		if mode in {'r', 'a'}:
			if not self.exists():
				raise FileNotFoundError(f"No such TarGit file '{self.filename!s}'")

			with tarfile.open(
					self.filename,
					mode="r:gz",
					format=tarfile.PAX_FORMAT,
					) as tf:
				check_archive_paths(tf)
				tf.extractall(path=self.__tmpdir_p)

			self.__repo = Repo(self.__tmpdir_p)
			self.__mode = mode
			self.__closed = False

		elif mode in {'w'}:
			if self.exists():
				raise FileExistsError(f"TarGit file '{self.filename!s}' already exists.")

			# Initialise git repo in tmpdir
			self.__repo = Repo.init(self.__tmpdir_p)
			self.__mode = mode
			self.__closed = False
			self.__do_commit(message="Empty initial commit.")

		else:
			raise ValueError(f"Unknown IO mode {mode!r}")

	def save(self) -> bool:
		"""
		Saves the contents of the archive.

		Does nothing if there are no changes to be saved.

		:returns: Whether there were any changes to save.

		:raises IOError: If the file is closed, or if it was opened in read-only mode.
		"""

		if self.closed:
			raise OSError("IO operation on closed TarGit file.")
		elif self.__mode not in {'w', 'a'}:
			raise OSError("Cannot write to TarGit file opened in read-only mode.")

		current_status = self.status()

		if any([
				current_status["add"] != [],
				current_status["delete"] != [],
				current_status["modify"] != [],
				]):
			# There are changes to commit
			message = "; ".join([
					f"{len(current_status['add'])} added",
					f"{len(current_status['delete'])} deleted",
					f"{len(current_status['modify'])} modified",
					])

			self.__do_commit(message)

			with self.filename.open("wb", buffering=False) as fp:
				with tarfile.open(
						self.filename,
						mode="w:gz",
						format=tarfile.PAX_FORMAT,
						fileobj=fp,
						) as tf:
					tf.add(str(self.__tmpdir_p), arcname='')

				fp.flush()

			return True
		return False

	def status(self) -> StagedDict:
		"""
		Returns the status of the TarGit archive.

		The values in the dictionary are lists of filenames, relative to the TarGit root.

		:raises IOError: If the file is closed.
		"""

		if self.closed:
			raise OSError("IO operation on closed TarGit file.")
		elif self.__mode not in {'w', 'a'}:
			return {"add": [], "delete": [], "modify": []}

		current_status = status(self.__tmpdir_p)

		for file in (*current_status.unstaged, *current_status.untracked):
			self.__repo.stage(str(file))

		return status(self.__tmpdir_p).staged

	def __do_commit(self, message: str):
		if self.closed:
			raise OSError("IO operation on closed TarGit file.")
		elif self.__mode not in {'w', 'a'}:
			raise OSError("Cannot write to TarGit file opened in read-only mode.")

		login = getpass.getuser()
		username = f"{login} <{login}@{socket.gethostname()}>"
		current_time = datetime.datetime.now(datetime.timezone.utc).astimezone()
		current_timezone = current_time.tzinfo.utcoffset(None).total_seconds()  # type: ignore

		self.__repo.do_commit(
				message=message.encode("UTF-8"),
				committer=username.encode("UTF-8"),
				author=username.encode("UTF-8"),
				commit_timestamp=current_time.timestamp(),
				commit_timezone=current_timezone,
				)

	def exists(self) -> bool:
		"""
		Returns whether the :class:`~.TarGit` archive exists.
		"""

		return self.filename.is_file()

	def close(self):
		"""
		Closes the :class:`~.TarGit` archive.
		"""

		self.__exit_handler()
		atexit.unregister(self.__exit_handler)

	def __exit_handler(self):
		if self.__tmpdir is not None:
			self.__tmpdir.cleanup()
		if self.__lock is not None:
			self.__lock.release()
		self.__closed = True

	@property
	def closed(self) -> bool:
		"""
		Returns whether the :class:`~.TarGit` archive is closed.
		"""

		return self.__closed

	@property
	def mode(self) -> Modes:
		"""
		Returns the mode the :class:`~.TarGit` archive was opened in.

		This defaults to ``'r'``. After the archive is closed this will show the
		last mode until the archive is opened again.
		"""

		return self.__mode

	def __truediv__(self, filename):
		"""
		Returns a :class:`~domdf_python_tools.paths.PathPlus` object
		representing the given filename relative to the archive root.

		:param filename:
		"""  # noqa: D400

		return self.__tmpdir_p / filename

	def __del__(self):
		self.close()

	def __repr__(self) -> str:
		"""
		Returns a string representation of the :class:`~.TarGit`.
		"""

		return f"{self.__class__.__name__}({self.filename})"

	def __fspath__(self) -> str:
		"""
		Returns the filename of the :class:`~.TarGit` archive.
		"""

		return os.fspath(self.filename)

	def __str__(self) -> str:
		"""
		Returns the filename of the :class:`~.TarGit` archive.
		"""

		return self.filename.as_posix()

	@property
	def history(self) -> Iterator[SaveState]:
		"""
		Returns an iterable over the historic save states of the :class:`~.TarGit`.
		:return:
		"""
		if self.closed:
			raise OSError("IO operation on closed TarGit file.")

		for entry in self.__repo.get_walker():
			# TODO: changed files

			author_m = re.match(r".*?\s+<(.*?)@(.*?)>", entry.commit.author.decode("UTF-8"))
			if author_m:
				user, device = author_m.groups()
			else:
				user, device = '', ''

			yield SaveState(
					id=entry.commit.id.decode("UTF-8"),
					user=user,
					device=device,
					time=entry.commit.author_time,
					timezone=entry.commit.author_timezone,
					)



#!/usr/bin/env python3
#
#  config.py
"""
Utilities for repository configuration.

.. versionadded:: 0.5.0
"""
#
#  Copyright © 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all
#  copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
#  DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
#  OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
#  OR OTHER DEALINGS IN THE SOFTWARE.
#
#  get_user_identity and Repo based on https://github.com/dulwich/dulwich
#  Copyright (C) 2013 Jelmer Vernooij <jelmer@jelmer.uk>
#  |  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  |  not use this file except in compliance with the License. You may obtain
#  |  a copy of the License at
#  |
#  |	  http://www.apache.org/licenses/LICENSE-2.0
#  |
#  |  Unless required by applicable law or agreed to in writing, software
#  |  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  |  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  |  License for the specific language governing permissions and limitations
#  |  under the License.
#

# stdlib
from typing import Dict

# 3rd party
from dulwich.config import ConfigFile

__all__ = ["set_remote_ssh", "set_remote_http", "get_remotes"]


def set_remote_ssh(config: ConfigFile, domain: str, username: str, repo: str, name: str = "origin"):
	"""
	Set the remote url for the repository, using SSH.

	:param config:
	:param domain:
	:param username:
	:param repo:
	:param name: The name of the remote to set.

	.. versionadded:: 0.5.0
	"""

	config.set(("remote", name), "url", f"git@{domain}:{username}/{repo}.git".encode("UTF-8"))


def set_remote_http(config: ConfigFile, domain: str, username: str, repo: str, name: str = "origin"):
	"""
	Set the remote url for the repository, using HTTP.

	:param config:
	:param domain:
	:param username:
	:param repo:
	:param name: The name of the remote to set.

	.. versionadded:: 0.5.0
	"""

	config.set(("remote", name), "url", f"https://{domain}/{username}/{repo}.git".encode("UTF-8"))


set_remote_html = set_remote_http


def get_remotes(config: ConfigFile) -> Dict[str, str]:
	"""
	Returns a dictionary mapping remote names to URLs.

	:param config:

	.. versionadded:: 0.5.0
	"""

	remotes = {}

	for key in list(config.keys()):
		if key[0] == b"remote":
			url = config.get(key, "url")
			if url is not None:
				remotes[key[1].decode("UTF-8")] = url.decode("UTF-8")

	return remotes
